// test_binsearch.cpp
// Glenn G. Chappell
// 5 Dec 2013
//
// For CS 202 Fall 2013
// Test program for function template binSearch

#include "binsearch.h"  // For function template binSearch
#include "binsearch.h"  // Double-inclusion test

#include <iostream>
using std::cout;
using std::endl;
using std::cin;
#include <cstdlib>
using std::size_t;
#include <deque>
using std::deque;
#include <string>
using std::string;


bool allGood = true;  // true if all tests pass so far


// bstr
// Return string form of bool: "true" or "false", as appropriate.
string bstr(bool b)
{
    return b ? "true" : "false";
}


// takesBool
// Function that takes bool by reference -- for return-type checking.
void takesBool(bool & b)
{}


// test_binSearch_int_small
// Test for binSearch: int values, small datasets
// Prints results of tests, sets global allGood to false if any fail
void test_binSearch_int_small()
{
    cout << endl;
    cout << "Test suite: int values, small datasets" << endl;

    // Make reverse-sorted const deque
    const size_t DATASIZE = 10;
    int data[DATASIZE] = { 100, 90, 80, 70, 60, 50, 40, 30, 20, 10 };
    const deque<int> d(data, data+DATASIZE);

    // Return-type check (passes if it compiles)
    {
    auto result = binSearch(d.begin(), d.end(), 20);
    takesBool(result);  // Should only compile if binSearch returns bool
    }

    // Empty ranges
    {
    int key = 50;
    auto first = d.rbegin()+3;
    auto last = first;
    bool result = binSearch(first, last, key);
    cout << "  Empty range #1 - should be false: "
         << bstr(result) << endl;;
    if (result)
    {
        cout << "    FAIL!!! ***********************************************"
             << endl;
        allGood = false;
    }
    }

    {
    int key = 50;
    auto first = d.rbegin()+4;
    auto last = first;
    bool result = binSearch(first, last, key);
    cout << "  Empty range #2 - should be false: "
         << bstr(result) << endl;
    if (result)
    {
        cout << "    FAIL!!! ***********************************************"
             << endl;
        allGood = false;
    }
    }

    {
    int key = 50;
    auto first = d.rbegin()+5;
    auto last = first;
    bool result = binSearch(first, last, key);
    cout << "  Empty range #3 - should be false: "
         << bstr(result) << endl;
    if (result)
    {
        cout << "    FAIL!!! ***********************************************"
             << endl;
        allGood = false;
    }
    }

    // Singleton ranges
    {
    int key = 40;
    auto first = d.rbegin()+2;
    auto last = first+1;
    bool result = binSearch(first, last, key);
    cout << "  Singleton range #1 - should be false: "
         << bstr(result) << endl;;
    if (result)
    {
        cout << "    FAIL!!! ***********************************************"
             << endl;
        allGood = false;
    }
    }

    {
    int key = 40;
    auto first = d.rbegin()+3;
    auto last = first+1;
    bool result = binSearch(first, last, key);
    cout << "  Singleton range #2 - should be true: "
         << bstr(result) << endl;
    if (!result)
    {
        cout << "    FAIL!!! ***********************************************"
             << endl;
        allGood = false;
    }
    }

    {
    int key = 40;
    auto first = d.rbegin()+4;
    auto last = first+1;
    bool result = binSearch(first, last, key);
    cout << "  Singleton range #3 - should be false: "
         << bstr(result) << endl;
    if (result)
    {
        cout << "    FAIL!!! ***********************************************"
             << endl;
        allGood = false;
    }
    }

    // Size 2
    {
    int key = 65;
    auto first = d.rbegin()+6;
    auto last = first+2;
    bool result = binSearch(first, last, key);
    cout << "  Size 2 range #1 - should be false: "
         << bstr(result) << endl;;
    if (result)
    {
        cout << "    FAIL!!! ***********************************************"
             << endl;
        allGood = false;
    }
    }

    {
    int key = 70;
    auto first = d.rbegin()+6;
    auto last = first+2;
    bool result = binSearch(first, last, key);
    cout << "  Size 2 range #2 - should be true: "
         << bstr(result) << endl;;
    if (!result)
    {
        cout << "    FAIL!!! ***********************************************"
             << endl;
        allGood = false;
    }
    }

    {
    int key = 75;
    auto first = d.rbegin()+6;
    auto last = first+2;
    bool result = binSearch(first, last, key);
    cout << "  Size 2 range #3 - should be false: "
         << bstr(result) << endl;;
    if (result)
    {
        cout << "    FAIL!!! ***********************************************"
             << endl;
        allGood = false;
    }
    }

    {
    int key = 80;
    auto first = d.rbegin()+6;
    auto last = first+2;
    bool result = binSearch(first, last, key);
    cout << "  Size 2 range #4 - should be true: "
         << bstr(result) << endl;;
    if (!result)
    {
        cout << "    FAIL!!! ***********************************************"
             << endl;
        allGood = false;
    }
    }

    {
    int key = 85;
    auto first = d.rbegin()+6;
    auto last = first+2;
    bool result = binSearch(first, last, key);
    cout << "  Size 2 range #5 - should be false: "
         << bstr(result) << endl;;
    if (result)
    {
        cout << "    FAIL!!! ***********************************************"
             << endl;
        allGood = false;
    }
    }

}


// test_binSearch_string_medium
// Test for binSearch: string values, medium-sized datasets
// Prints results of tests, sets global allGood to false if any fail
void test_binSearch_string_medium()
{
    cout << endl;
    cout << "Test suite: string values, medium-sized datasets" << endl;

    // Make reverse-sorted const deque
    const size_t DATASIZE = 10;
    string data[DATASIZE] = { "zither", "zebra", "whig",
                              "putrescence", "murgatroyd",
                              "incomprehensibilities", "h",
                              "glug", "boat", "ant" };
    const deque<string> d(data, data+DATASIZE);

    // Return-type check (passes if it compiles)
    {
    auto result = binSearch(d.begin(), d.end(), "goat");
    takesBool(result);  // Should only compile if binSearch returns bool
    }

    {
    string key = "whig";
    auto first = d.rbegin()+1;
    auto last = d.rend()-1;
    bool result = binSearch(first, last, key);
    cout << "  Key in dataset #1 - should be true: "
         << bstr(result) << endl;
    if (!result)
    {
        cout << "    FAIL!!! ***********************************************"
             << endl;
        allGood = false;
    }
    }

    {
    string key = "boat";
    auto first = d.rbegin()+1;
    auto last = d.rend()-1;
    bool result = binSearch(first, last, key);
    cout << "  Key in dataset #2 - should be true: "
         << bstr(result) << endl;
    if (!result)
    {
        cout << "    FAIL!!! ***********************************************"
             << endl;
        allGood = false;
    }
    }

    {
    string key = "nnn";
    auto first = d.rbegin()+1;
    auto last = d.rend()-1;
    bool result = binSearch(first, last, key);
    cout << "  Key within range of values, not present - should be false: "
         << bstr(result) << endl;
    if (result)
    {
        cout << "    FAIL!!! ***********************************************"
             << endl;
        allGood = false;
    }
    }

    {
    string key = "ant";
    auto first = d.rbegin()+1;
    auto last = d.rend()-1;
    bool result = binSearch(first, last, key);
    cout << "  Key before range of values - should be false: "
         << bstr(result) << endl;
    if (result)
    {
        cout << "    FAIL!!! ***********************************************"
             << endl;
        allGood = false;
    }
    }

    {
    string key = "zither";
    auto first = d.rbegin()+1;
    auto last = d.rend()-1;
    bool result = binSearch(first, last, key);
    cout << "  Key after range of values - should be false: "
         << bstr(result) << endl;
    if (result)
    {
        cout << "    FAIL!!! ***********************************************"
             << endl;
        allGood = false;
    }
    }
}


// test_binSearch_int_large
// Test for binSearch: int values, large datasets
// Prints results of tests, sets global allGood to false if any fail
void test_binSearch_int_large()
{
    cout << endl;
    cout << "Test suite: int values, large datasets" << endl;

    // Make reverse-sorted const deque
    cout << "  CONSTRUCTING DATASET ... ";
    cout.flush();
    const size_t DATASIZE = 20000000;  // REDUCE THIS, IF TOO SLOW
    deque<int> d;
    for (size_t i = 0; i < DATASIZE; ++i)
        d.push_front(int(i) * 2);
    cout << "DONE" << endl;

    {
    int key = (DATASIZE/3)*2;
    auto first = d.rbegin()+1;
    auto last = d.rend()-1;
    bool result = binSearch(first, last, key);
    cout << "  Key in dataset #1 - should be true: "
         << bstr(result) << endl;
    if (!result)
    {
        cout << "    FAIL!!! ***********************************************"
             << endl;
        allGood = false;
    }
    }

    {
    int key = (DATASIZE-2)*2;
    auto first = d.rbegin()+1;
    auto last = d.rend()-1;
    bool result = binSearch(first, last, key);
    cout << "  Key in dataset #2 - should be true: "
         << bstr(result) << endl;
    if (!result)
    {
        cout << "    FAIL!!! ***********************************************"
             << endl;
        allGood = false;
    }
    }

    {
    int key = (DATASIZE/3*2)*2+1;
    auto first = d.rbegin()+1;
    auto last = d.rend()-1;
    bool result = binSearch(first, last, key);
    cout << "  Key within range of values, not present - should be false: "
         << bstr(result) << endl;
    if (result)
    {
        cout << "    FAIL!!! ***********************************************"
             << endl;
        allGood = false;
    }
    }

    {
    int key = 0;
    auto first = d.rbegin()+1;
    auto last = d.rend()-1;
    bool result = binSearch(first, last, key);
    cout << "  Key before range of values - should be false: "
         << bstr(result) << endl;
    if (result)
    {
        cout << "    FAIL!!! ***********************************************"
             << endl;
        allGood = false;
    }
    }

    {
    int key = (DATASIZE-1)*2;
    auto first = d.rbegin()+1;
    auto last = d.rend()-1;
    bool result = binSearch(first, last, key);
    cout << "  Key after range of values - should be false: "
         << bstr(result) << endl;
    if (result)
    {
        cout << "    FAIL!!! ***********************************************"
             << endl;
        allGood = false;
    }
    }
}




// main program
// Runs tests for binSearch, prints results
int main()
{
    cout << "Test function template binSearch" << endl;
    test_binSearch_int_small();
    test_binSearch_string_medium();
    test_binSearch_int_large();
    
    cout << endl;
    if (allGood)
        cout << "All tests pass!" << endl;
    else
        cout << "***** SOME TESTS FAILED *****" << endl;

    cout << endl;
    cout << "Press ENTER to quit ";
    while (cin.get() != '\n') ;

    return 0;
}

